import os
from elasticsearch import Elasticsearch, helpers
import json


def create_index(es, index):
	if es.indices.exists(index):
		print("deleting '%s' index..." % index)
		res = es.indices.delete(index=index)
		print(" response: '%s'" % res)

	mapping = {
		"settings": {
			"number_of_shards": 1,
			"number_of_replicas": 0
		},
		"mappings": {
			'email': {
				"properties": {
					"quagga.parsed": {
						"type": "object",
						"properties": {
							"blocks": {
								"type": "nested",
								"properties": {
									"from": {
										"type": "object",
										"properties": {
											"name": {
												"type": "keyword",
											},
											"email": {
												"type": "keyword"
											}

										}
									},
									"raw_from": {
										"type": "text"
									},
									"to": {
										"type": "nested",
										"properties": {
											"name": {
												"type": "keyword",
											},
											"email": {
												"type": "keyword"
											}
										}
									},
									"raw_to": {
										"type": "text"
									},
									"cc": {
										"type": "nested",
										"properties": {
											"name": {
												"type": "keyword",
											},
											"email": {
												"type": "keyword"
											}
										}
									},
									"raw_cc": {
										"type": "text"
									},
									"sent": {
										"type": "date",
										"format": "yyyy-MM-dd HH:mm:ss z"
									},
									"raw_sent": {
										"type": "text"
									},
									"subject": {
										"type": "text"
									},
									"raw_subject": {
										"type": "text"
									},
									"type": {
										"type": "keyword"
									},
									"raw_header": {
										"type": "text"
									},
								}
							}
						}
					}
				}

			}
		}
	}

	print("creating '%s' index..." % (index))
	res = es.indices.create(index=index, body=mapping)
	print(" response: '%s'" % (res))


def gen_actions(es_index):
	def wrapper(item, id):

		for block in item['quagga.parsed']['blocks']:

			block['from'] = {
				'name': '',
				'email': '',
				'raw_name': ''
			} if block['from'] is None else block['from']
			for t in ['to', 'cc']:
				block[t] = [{
					'name': '',
					'email': '',
					'raw_name': ''
				}] if block[t] is None else block[t]
			for t in ['raw_cc', 'raw_to', 'raw_from', 'raw_sent', 'subject', 'raw_subject']:
				block[t] = '' if block[t] is None else block[t]
			block["sent"] = '2000-01-01 00:00:00 UTC' if block["sent"] is None or block["sent"] == "" else block["sent"]

		return {
			'_op_type': 'index',
			'_index': es_index,
			'_type': 'email',
			'_id': id,
			'_source': item
		}

	return wrapper


def gen_data(prep, dir):
	for i, file in enumerate(os.listdir(os.fsencode(dir))):
		filename = os.fsdecode(file)
		if filename.endswith(".json"):
			with open(os.path.join(dir, filename), "r") as f:
				print('Uploading: {}'.format(filename))
				yield prep(json.loads(f.read()), i)


def load(es, index):
	buffer_size = 100
	dir = "output/parsed"

	create_index(es, index)

	for ok, response in helpers.streaming_bulk(client=es,
	                                          actions=gen_data(gen_actions(index), dir),
	                                          chunk_size=buffer_size,
	                                          raise_on_exception=False,
	                                          raise_on_error=False,
	                                          max_retries=2,
	                                          yield_ok=False):
		if not ok:
			with open("uploadlog.txt", "w") as f:
				f.write(json.dumps(response) + "\n")


"""
def search(es, index):
	query = {"query": {"nested": {"path": "quagga.parsed.blocks", "score_mode": "avg", "query": {"bool": {"must": [{"match": {"quagga.parsed.blocks.from.name": "mark.haedicke"}},]}}}}}


	res = es.search(index=index, body=query)
	print(res)"""

if __name__ == '__main__':
	es = Elasticsearch([{'host': 'localhost', 'port': 9200}])
	index = "quagga_enron"

	load(es, index)
