

package org.benhurdelhey

import org.apache.spark.graphx.Edge
import org.apache.spark.graphx.Graph
import org.apache.spark.graphx.GraphLoader
import org.apache.spark.graphx.VertexId
import org.apache.spark.graphx.PartitionStrategy
import org.apache.spark.graphx.util.GraphGenerators

import org.apache.spark.rdd.RDD

import org.apache.spark.sql.SparkSession
import org.apache.spark.SparkContext

import org.json4s._
import org.json4s.jackson.JsonMethods._

import scala.collection.mutable.ListBuffer
import scala.util.Random

import java.util.Calendar
import java.io.PrintWriter

case class Person(name: String, email: String, raw_name: String)
case class Block(from: Person, to: List[Person], cc: List[Person])

case class Triplet (srcGroup : Long , dstGroup : Long , weight : Double , rank : Long, originalSrc : Long, originalDst : Long)


/*TODO
- adjust mincut definition
- read from config so its easier to determine constants
- pagerank algorithm
*/

/*
object RandomizedMinCut {

  private var g : Graph[Long, Triplet] = null
  private var checkpoint_interval : Long = 200
  private var tripletRankOrdering : Ordering[Triplet] = new Ordering[Triplet]() {
      override def compare(x: Triplet, y: Triplet): Int = 
          Ordering[Long].compare(x.rank, y.rank)
    }

  private def randomEdgeRank(): Long = {
    val rand = new Random()
    rand.nextLong()
  }
  
  private def contractEdge(edges: RDD[Triplet]): RDD[Triplet] = {
    val max_rank = edges.maxBy()(tripletRankOrdering)
    val oldGroup = math.max(max_rank.srcGroup , max_rank.dstGroup)
    val newGroup = math.min(max_rank.srcGroup , max_rank.dstGroup) 
    edges
      .map(e => 
        if (e.dstGroup == oldGroup) 
          e.copy(dstGroup=newGroup)
        else if (e.srcGroup == oldGroup) 
          e.copy(srcGroup=newGroup)
        else e)
      .filter(e => (e.srcGroup != e.dstGroup))
  }

  private def brute_force_mincut(triplets: RDD[Triplet], numVertices : Long): (Double, RDD[Triplet]) = {
    var edges = triplets
    var components = numVertices // first stage contractEdgeions
    while (components > 2) {
      ////println("components " + components)
      ////println(edges.collect().mkString("\n"))
      edges = contractEdge(edges) 
      components = components - 1
      if (( components % checkpoint_interval) == 0) {
        edges.checkpoint()
      }
    }

    // sum the weight for edges which are a candidate for the mincut
    (edges.map(e => e.weight).sum, edges)
  }


  private def karger_stein(orig_edges : RDD[Triplet], numVertices : Long) : (Double, RDD[Triplet]) = {
    if (numVertices < 6) {
      brute_force_mincut(orig_edges , numVertices)
    } else {
      val t = math.ceil(numVertices / math.sqrt(2)).toInt 
      var edges1 : RDD[Triplet] = orig_edges.map(e => e.copy(rank=randomEdgeRank()) )
      var edges2 : RDD[Triplet] = orig_edges.map(e => e.copy(rank=randomEdgeRank()) )
      edges1.cache
      edges2.cache
      var components = numVertices
      // first stage contractEdgeions 
      while (components > t) {
        edges1 = contractEdge(edges1) 
        edges2 = contractEdge(edges2) 
        components = components - 1
        if (( components % checkpoint_interval) == 0) {
          edges1.checkpoint
          edges2.checkpoint
        } 
      }
      val res1 = karger_stein(edges1, components)
      val res2 = karger_stein(edges2, components)
      if (res1._1 < res2._1)
        res1
      else
        res2
    } 
  }

  def run(sc: SparkContext, triplets : RDD[Triplet], numVertices : Long, trials : Long): (Double, RDD[Triplet], RDD[Triplet]) = {
    var min_cut: (Double, RDD[Triplet]) = (Double.PositiveInfinity, sc.emptyRDD[Triplet])
    val begin : Long = 1
    triplets.cache()

    begin to trials foreach { _ =>
      val cut = karger_stein(triplets, numVertices)
      min_cut = if (cut._1 < min_cut._1) cut else min_cut}
    
    if (min_cut._2.count == 0) {
      //println("no min cut found")
    }

    val new_triplets = triplets
    .subtract(min_cut._2
      .map(triplet => triplet
        .copy(rank=0, srcGroup=triplet.originalSrc, dstGroup=triplet.originalDst)))

    (min_cut._1, min_cut._2, new_triplets)
  }

}*/



object ScalaRandomizedMinCut {


  private def randomEdgeRank(): Long = {
    val rand = new Random()
    rand.nextLong()
  }
  
  private def contractEdge(edges: Iterable[Triplet]): Iterable[Triplet] = {
    val max_rank = edges.maxBy(e => e.rank)
    val oldGroup = math.max(max_rank.srcGroup , max_rank.dstGroup)
    val newGroup = math.min(max_rank.srcGroup , max_rank.dstGroup) 
    edges
      .map(e => 
        if (e.dstGroup == oldGroup) 
          e.copy(dstGroup=newGroup)
        else if (e.srcGroup == oldGroup) 
          e.copy(srcGroup=newGroup)
        else e)
      .filter(e => (e.srcGroup != e.dstGroup))
  }

  private def brute_force_mincut(triplets: Iterable[Triplet], numVertices : Long): (Double, Iterable[Triplet]) = {
    var edges = triplets
    var components = numVertices // first stage contractEdgeions
    while (components > 2) {
      ////println("components " + components)
      ////println(edges.collect().mkString("\n"))
      edges = contractEdge(edges) 
      components = components - 1
    }

    // sum the weight for edges which are a candidate for the mincut
    (edges.map(e => e.weight).sum, edges)
  }


  private def karger_stein(orig_edges : Iterable[Triplet], numVertices : Long) : (Double, Iterable[Triplet]) = {
    if (numVertices < 6) {
      brute_force_mincut(orig_edges , numVertices)
    } else {
      val t = math.ceil(numVertices / math.sqrt(2)).toInt 
      var edges1 : Iterable[Triplet] = orig_edges.map(e => e.copy(rank=randomEdgeRank()) )
      var edges2 : Iterable[Triplet] = orig_edges.map(e => e.copy(rank=randomEdgeRank()) )
      var components = numVertices

      while (components > t) {
        edges1 = contractEdge(edges1) 
        edges2 = contractEdge(edges2) 
        components = components - 1
      }
      val res1 = karger_stein(edges1, components)
      val res2 = karger_stein(edges2, components)
      if (res1._1 < res2._1)
        res1
      else
        res2
    } 
  }

  def run(triplets : Iterable[Triplet], numVertices : Long, trials : Long): (Double, Iterable[Triplet]) = {
    var min_cut: (Double, Iterable[Triplet]) = (Double.PositiveInfinity, new Array[Triplet](0))
    val begin : Long = 1

    begin to trials foreach { i =>
      println("mincut trial" + i + " at " + Calendar.getInstance().getTime() + " of " + trials)
      val cut = karger_stein(triplets, numVertices)
      min_cut = if (cut._1 < min_cut._1) cut else min_cut}
    
    if (min_cut._2.size == 0) {
      //println("no min cut found")
    }

    (min_cut._1, min_cut._2)
  }

}


object ScalaPageRank {

  def run(links: Map[Long, Iterable[Long]]): Iterable[(Long, Double)] = {
    val iters = 10

    var ranks: Iterable[(Long, Double)] = links.toSeq.map(v => (v._1, 1.0))

    for (i <- 1 to iters) {

      var list: ListBuffer[(Iterable[Long], Double)] = ListBuffer[(Iterable[Long], Double)]()
      val contribs: Iterable[(Long, Double)] = ranks
        .foldLeft(list){(acc, b) => 
          links.get(b._1) match {
            case Some(u) =>
              val c: (Iterable[Long], Double) = (u, b._2)
              acc += c
            case None => //println("error no match")
          }
          
          acc}
        .flatMap{ case (urls, rank) =>
          val size = urls.size
          urls.map(url => (url, rank / size))}



      val defMap = contribs.map(v => (v._1, List[Double]())).toMap
      val newRanks: Iterable[(Long, Iterable[Double])] = contribs
        .foldLeft{defMap}(
          (acc, x) => acc + (x._1 -> (acc(x._1) :+ x._2)))
        .toSeq
      ranks = newRanks
        .map(v => (v._1, v._2.sum))
        .map(v => (v._1, 0.15 + 0.85 * v._2))
    }

    ranks
  }
}

object ConnectedComponents {

  def run(adjMap: Map[Long, List[Long]]): List[List[Long]] = {
      val v = adjMap.keys
      var globalVisits = List[Long]()
      def BFS_r(elems: List[Long], visited: List[List[Long]]): Option[List[List[Long]]] =
      {
          val newNeighbors = elems.flatMap(adjMap(_)).filterNot(visited.flatten.contains).distinct
          if (newNeighbors.isEmpty)
              Some(visited)
          else
              BFS_r(newNeighbors, newNeighbors :: visited)
      }
      v.flatMap(x => {
          if (globalVisits.contains(x))
              None
          else
          {
              val vi: List[Long] = BFS_r(List(x), List(List(x))).get.flatten
              globalVisits = globalVisits ++ vi
              Some(vi)
          }
      }).toList
  }
}






object PersonDeduplicator {

  private var sc: SparkContext = null
  private var parsed_dir: String = ""
  private var checkpoint_dir: String = ""
  private var output_dir: String = ""
  private var graph: Graph[String, Double] = null
  private var cc: RDD[(VertexId, VertexId)] = null

  private def parseArgs(args: Array[String]) : Unit = {
    //println(args)

    parsed_dir = args(0)
    //println("PARSED DIR: " + parsed_dir)

    checkpoint_dir = args(1)
    //println("CHECKPOINT DIR: " + checkpoint_dir)

    output_dir = args(2)
    //println("OUTPUT DIR: " + output_dir)
  }

  private def toGexf[VD,ED](g:Graph[VD,ED]) : String = {
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
    "<gexf xmlns=\"http://www.gexf.net/1.2draft\" version=\"1.2\">\n" +
    "  <graph mode=\"static\" defaultedgetype=\"directed\">\n" +
    "    <nodes>\n" +
    g.vertices.map{ v =>
      "      <node id=\"" + v._1 + "\" label=\"" +
                        v._2.asInstanceOf[String].replace("&", " ").replace("<", " ").replace(">", " ").replace("'", " ").replace("\"", " ") + "\" />\n"}
      .collect.mkString +
    "    </nodes>\n" +
    "    <edges>\n" +
    g.edges.map( e => "      <edge source=\"" + e.srcId +
                     "\" target=\"" + e.dstId + "\" weight=\"" + e.attr +
                     "\" />\n").collect.mkString +
    "    </edges>\n" +
    "  </graph>\n" +
    "</gexf>"
  }

  private def createGraph() : Unit = {

    val all_names = sc.textFile(parsed_dir)
      .coalesce(64)
      .flatMap { file => 
        implicit val formats = DefaultFormats // Brings in default date formats etc.
        val json = parse(file)
        ((json \ "quagga.parsed") \ "blocks")
          .extract[List[Block]]
          .zipWithIndex
          .flatMap { case (block, i) =>
            var weight: Double = 1.0
            if (i == 0) {
              weight = Double.PositiveInfinity
            }
            val names: ListBuffer[((String, String), Double)] = ListBuffer(((block.from.email, block.from.name), weight))
            names ++= block.to.map{ name => ((name.email, name.name), weight)}
            names ++= block.cc.map{ name => ((name.email, name.name), weight)}
            names}}
      .reduceByKey(_ + _)
      .cache

    val filtered_names = all_names
      .filter { case ((email, name), _) =>
        !email.isEmpty && !name.isEmpty && (6 < name.length() && name.length() <= 50 && name.length() <= 50)}

    val name_ids = all_names
        .flatMap {name => Array((name._1._1), (name._1._2))}
        .distinct
        .zipWithUniqueId
        .cache

    val vertices: RDD[(VertexId, String)] = name_ids
        .map{ case (name, id) => (id.toLong, name)}

    val edges = filtered_names
        .map{name => (name._1._1, (name._1._2, name._2))}
        .join(name_ids).map {name => (name._2._1, name._2._2)}
        .map{name => (name._1._1, (name._2, name._1._2))} //last is count
        .join(name_ids).map {name => (name._2._1, name._2._2)}
        .map{name => ((name._1._1, name._2), name._1._2)}
        .map{edge => Edge(edge._1._1, edge._1._2, edge._2)}
 
    graph = Graph.apply(vertices, edges, "")
    graph.partitionBy(PartitionStrategy.EdgePartition2D)
  }

  private def store_gexf(i: String, g: Graph[String, Double]): Unit = {
    new PrintWriter(output_dir + "graph" + i + ".gexf") { write(toGexf(g)); close }
  }

  private def store_cc(): Unit = {
    val ccByUsername = graph.vertices.join(cc)
      .map { case (_, (username, ccId)) => (username, ccId) }
      .saveAsTextFile(output_dir)
  }

  private def determineNewCc(ccVertices: Iterable[Long], newCcTriplets: Iterable[Triplet]): Iterable[Iterable[Long]] = {
    val defMap = ccVertices.map(_ -> List[Long]()).toMap
    val adjMap = newCcTriplets.foldLeft{defMap}(
        (acc, x) => acc + (x.originalSrc -> (acc(x.originalSrc) :+ x.originalDst)))
    ////println("adjmap \n" + adjMap)
    val newCc = ConnectedComponents.run(adjMap)
    newCc
  }

  private def shouldPerformMinCut(ccTriplets: Iterable[Triplet], ccVertices: Iterable[Long], mincut: (Double, Iterable[Triplet])): Boolean = {
    // evaluate mincut
    val newCcTriplets = (ccTriplets.toSet --
      (mincut._2.map(triplet => triplet
        .copy(rank=0, srcGroup=triplet.originalSrc, dstGroup=triplet.originalDst)).toSet).toSeq)
      .flatMap{t => Array(t, t.copy(originalSrc = t.originalDst, originalDst = t.originalSrc))}
    val newCc = determineNewCc(ccVertices, newCcTriplets)
    //println("new cc \n" + newCc)

    val smallestNewCcSize = newCc.map(cc => cc.size).min
    val evenCutProb = (2 * (smallestNewCcSize / ccTriplets.size)) // in [0, 1]
    
    val avgCcEdgeWeights = newCc
      .map{cc =>cc
          .flatMap{vId => newCcTriplets.filter(t => t.originalSrc == vId)}.map(t => t.weight)} 
      .map{cc => cc.sum / cc.size}
    val avgCcEdgeWeight = avgCcEdgeWeights.sum / avgCcEdgeWeights.size
    val cutEdgeWeights = mincut._2.map(t => t.weight)
    val avgCutEdgeWeight = cutEdgeWeights.sum / cutEdgeWeights.size
    val edgeWeightProb = 1 - (avgCutEdgeWeight / avgCcEdgeWeight)

    val ccSizeProb = Math.max(1.0, Math.pow(ccVertices.size / 17, 2))

    val cutProb = (evenCutProb + edgeWeightProb + 2 * ccSizeProb) / 4

    cutProb > 0.4
  }

  private def performCutIteration(ccTriplets: Iterable[Triplet], ccVertices: Iterable[Long]): (Iterable[Iterable[Triplet]], Iterable[Triplet]) = {

    val trials: Long = math.pow(math.log(ccVertices.size.toDouble), 2).toLong

    var cutCcTripletsForIteration = collection.mutable.Set[Triplet]()






    // PAGERANK

    val defMap = ccVertices.map(_ -> List[Long]()).toMap
    val links: Map[Long, Iterable[Long]] = ccTriplets
      .flatMap{t => Array(t, t.copy(originalSrc=t.originalDst, originalDst=t.originalSrc))}
      .map(t => (t.originalSrc, t.originalDst))
      .toSet.toSeq
      .foldLeft{defMap}(
        (acc, x) => acc + (x._1 -> (acc(x._1) :+ x._2)))

    val ccVerticesWithPagerank = ScalaPageRank.run(links)
    //println("pageranks\n" + ccVerticesWithPagerank)

    val vertexWithHighestPagerank = ccVerticesWithPagerank
      .maxBy(v => v._2)
    //println("=====found vertex with highest pagerank " + vertexWithHighestPagerank)

    val avgPagerank = ccVerticesWithPagerank
      .map{case(vId, pagerank) => (pagerank)}
      .sum / ccVertices.size
    //println("avg pagerank " + avgPagerank)

    // todo: determine constants
    val numOtherPopularVertices = ccVerticesWithPagerank
      .filter{case(vId, pagerank) => vertexWithHighestPagerank._2 * 0.6 <= pagerank}
      .size - 1
    //println("number of other vertices with high pagerank: " + numOtherPopularVertices)

    if (avgPagerank < 0.1 * vertexWithHighestPagerank._2 &&
      (numOtherPopularVertices / ccVertices.size) < 0.1) {
      ////println("splitting edges at max pagerank vertex")

      var pagerankCutTriplets = ccTriplets
        .filter(t => t.originalSrc == vertexWithHighestPagerank._1 ||
          t.originalDst == vertexWithHighestPagerank._1)

      val highestWeightTriplet = pagerankCutTriplets
        .maxBy(v => v.weight)
      
      pagerankCutTriplets = pagerankCutTriplets
        .filter(t => t != highestWeightTriplet)

      println("pagerank cut")

      cutCcTripletsForIteration = cutCcTripletsForIteration ++ pagerankCutTriplets
    }




    
    // MINCUT
    // todo: alter mincut definition so its more likely to choose such mincut we search for
    val mincut = ScalaRandomizedMinCut.run(ccTriplets, ccVertices.size, trials)
    //println("mincut in component " + cc + " : " + mincut)

    if (mincut._2.size > 0 && shouldPerformMinCut(ccTriplets, ccVertices, mincut)) {
      cutCcTripletsForIteration = cutCcTripletsForIteration ++ mincut._2
      println("mincut")
    }


    val newTriplets = ccTriplets.toSet --
      (cutCcTripletsForIteration.map(triplet => triplet
        .copy(rank=0, srcGroup=triplet.originalSrc, dstGroup=triplet.originalDst)).toSet).toSeq      

    val newCc = determineNewCc(ccVertices, newTriplets)

    val newTripletsByNewCc = newCc.map{cc =>
      cc.flatMap{vId => newTriplets.filter(t => t.originalSrc == vId)}}

    (newTripletsByNewCc, cutCcTripletsForIteration)
  }

  private def cutCc(res: Iterable[Triplet], maxRecursionDepth: Long): collection.mutable.Set[Triplet] = {
    var cutCcTriplets = collection.mutable.Set[Triplet]()
    var ccTriplets = res
    var ccVertices = ccTriplets
      .flatMap(t => Array(t.originalSrc, t.originalDst))
      .toSet
      .toIterable

    if (ccVertices.size > 3 && maxRecursionDepth > 0) {
      //println("triplets " + ccTriplets)
      //println("vertices " + ccVertices)

      performCutIteration(ccTriplets, ccVertices) match {
        case (newCcTriplets: Iterable[Iterable[Triplet]], cutCcTripletsForIteration) =>
          cutCcTriplets = cutCcTriplets ++ cutCcTripletsForIteration
          
          if (cutCcTripletsForIteration.size != 0) {
            newCcTriplets.foreach{ newCcTriplets =>
              cutCcTriplets = cutCcTriplets ++ cutCc(newCcTriplets, maxRecursionDepth - 1)
            }
          }
      }

    }

    cutCcTriplets.map(t => t.copy(rank = 0, srcGroup = t.originalSrc, dstGroup = t.originalDst))
  }

  def main(args: Array[String]): Unit = {

    val spark = SparkSession
      .builder
      .appName(s"${this.getClass.getSimpleName}")
      .config("spark.memory.fraction", "0.8")
      .getOrCreate

    parseArgs(args)

    sc = spark.sparkContext
    sc.setLogLevel("ERROR")
    sc.setCheckpointDir(checkpoint_dir)
    
    createGraph()

    store_gexf("original", graph)

    var finishedCc: Set[Long] = Set[Long]()

    1 to 100 foreach { i =>

      println("iteration " + i)

      cc = graph.connectedComponents.vertices

      val workingGraph = graph
          .mapTriplets(e => Triplet(e.srcId, e.dstId, e.attr, 0, e.srcId, e.dstId)) 
          .subgraph(epred = (triplet) => (triplet.srcId != triplet.dstId))

      val allTriplets = workingGraph
        .edges
        .map(e => e.attr)

      val tripletsByCc: RDD[(VertexId, Iterable[Triplet])] = allTriplets
        .map{t: Triplet => (t.originalSrc, t)}
        .join(cc)
        .map{case (vId, (triplet, ccId)) => (ccId, triplet)}
        .groupByKey
        .filter(cc => !finishedCc.contains(cc._1))

      //println(tripletsByCc.collect.mkString("\n"))

      val broadcastTripletsByCc = sc.broadcast(tripletsByCc.collect)

      val result: RDD[(Long, collection.mutable.Set[Triplet], Boolean)] = tripletsByCc
          .coalesce(64)
          .map{case (cc, _) =>
            val res = broadcastTripletsByCc
              .value
              .find{case(component, triplets) => component == cc}
            var cutTripletsForCc = collection.mutable.Set[Triplet]()
            res match {
              case None =>
                //println("No name value")

              case Some(t) =>
                val triplets = t._2
                cutTripletsForCc = cutCc(triplets, 8)
            }
            val ct = cutTripletsForCc.map(t => t.copy(rank = 0, srcGroup = t.originalSrc, dstGroup = t.originalDst))
            (cc, ct, ct.size == 0)
          }

      val cutTriplets = result.map(r => r._2)
      val iterationFinishedCc: Set[Long] = result.map(r => (r._1, r._3)).filter(r => r._2).map(r => r._1).collect.toSet
      finishedCc = finishedCc ++ iterationFinishedCc

      val allCutTriplets = cutTriplets.fold(collection.mutable.Set[Triplet]())((acc, b) => acc ++ b)
      println("all triplets to cut\n" + allCutTriplets.toSeq)

      graph = graph.subgraph(epred = (triplet) => {
          val t = Triplet(triplet.srcId, triplet.dstId, triplet.attr, 0, triplet.srcId, triplet.dstId)
          val r: Boolean = !allCutTriplets.contains(t)
          r}
      )
        

    }
    
    store_gexf("result", graph)
/*
    

    do {
      //println("iteration " + iteration.toString)
      

      cc = workingGraph
        .connectedComponents
        .vertices

      // make sure its bidirectional because pagerank takes a directed graph
      val verticesWithPagerank: RDD[(VertexId, Double)] = Graph.fromEdges(workingGraph
          .edges
          .map{e => Edge(e.srcId, e.dstId, 1)}
          .flatMap{e => Array(e, Edge(e.dstId, e.srcId, 1))}, "")
        .pageRank(0.0001)
        .vertices

      ////println("===verticesWithPagerank\n" + verticesWithPagerank.collect().mkString("\n"))

      val tripletsByCc: RDD[(VertexId, Iterable[Triplet])] = workingGraph
        .edges
        .map(e => e.attr)
        .map{t: Triplet => (t.originalSrc, t)}
        .join(cc)
        .map{case (vId, (triplet, ccId)) => (ccId, triplet)}
        .groupByKey

      var allNewTriplets: RDD[Triplet] = sc.emptyRDD[Triplet]
      var allCutTriplets: RDD[Triplet] = sc.emptyRDD[Triplet]

      ////println("===verticsWithNames\n" + graph.vertices.collect().mkString("\n"))
      ////println("===components\n" + tripletsByCc.collect.mkString("\n"))

      tripletsByCc
        .map{case (cc, _) =>
          val ccTriplets = collectedTriplets.filter((component, triplets) => component == cc)
        }
      // todo: map 
      tripletsByCc
        .map{case (cc, triplets) => (cc)}
        .collect
        .foreach{(cc) =>

          val ccTriplets = tripletsByCc
            .filter{case(ccCandidate, triplets) => cc == ccCandidate}
            .flatMap{case(cc, triplets) => triplets}
          val ccVertices: RDD[VertexId] = ccTriplets
              .flatMap(t => Array((t.originalSrc), (t.originalDst)))
              .distinct
          val numVertices = ccVertices.count

          if (numVertices > 3) { //makes no sense to split otherwise..
            ////println("==== component " + cc)
            /*
            // MINCUT
            val trials: Long = math.pow(math.log(numVertices.toDouble), 2).toLong
            ////println("NUM MINCUT TRIALS " + trials)
            val result = RandomizedMinCut.run(sc, ccTriplets, ccTriplets.count, trials)
            result match {
              case (count: Double, cutTriplets : RDD[Triplet], newTriplets: RDD[Triplet]) =>
                // todo: create new graph for this cc, determine the new two ccs and evaluate whether 
                // we want to split
                
                ////println("=====NEW TRIPLETS\n" + newTriplets.collect().mkString("\n"))
                ////println("=====CUT TRIPLETS\n" + cutTriplets.collect().mkString("\n"))
               
                // todo evaluate cut and add cut triplets if necessary


                allNewTriplets = allNewTriplets ++ newTriplets
                allCutTriplets = allCutTriplets ++ cutTriplets
            }*/

            // PAGERANK CUT
            val ccVerticesWithPagerank = ccVertices
              .map{v => (v, 0)}
              .join(verticesWithPagerank)
              .map{case(vId, (_, pagerank)) => (vId, pagerank)}

            val vertexWithHighestPagerank = ccVerticesWithPagerank
              .max()(new Ordering[(VertexId, Double)]() {
                  override def compare(x: (VertexId, Double), y: (VertexId, Double)): Int = 
                    Ordering[Double].compare(x._2, y._2)})
            ////println("=====found vertex with highest pagerank " + vertexWithHighestPagerank)

            val avgPagerank = ccVerticesWithPagerank
              .map{case(vId, pagerank) => (pagerank)}
              .sum / numVertices
            ////println("avg pagerank " + avgPagerank)

            // todo: determine constants
            val numOtherPopularVertices = ccVerticesWithPagerank
              .filter{case(vId, pagerank) => vertexWithHighestPagerank._2 * 0.7 <= pagerank}
              .count - 1
            ////println("number of other vertices with high pagerank: " + numOtherPopularVertices)

            if (avgPagerank < 0.3 * vertexWithHighestPagerank._2 &&
              (numOtherPopularVertices / numVertices) < 0.1) {
              ////println("splitting edges at max pagerank vertex")

              var cutTriplets = ccTriplets
                .filter(t => t.originalSrc == vertexWithHighestPagerank._1 ||
                  t.originalDst == vertexWithHighestPagerank._1)

              val highestWeightTriplet = cutTriplets
                .max()(new Ordering[Triplet]() {
                  override def compare(x: Triplet, y: Triplet): Int = 
                    Ordering[Long].compare(x.rank, y.rank)})
              
              cutTriplets = cutTriplets
                .filter(t => t != highestWeightTriplet)

              allCutTriplets = allCutTriplets ++ cutTriplets
            }

          }

        }

      if (allCutTriplets.count == 0) {
        converged = true
      } else {
        val cutTripletSeq = allCutTriplets.collect()
        graph = graph.subgraph(epred = (triplet) =>
          cutTripletSeq.filter(t => t.originalSrc == triplet.srcId && t.originalDst == triplet.dstId)
            .length == 0)
      
        store_gexf(iteration.toString, graph)
        iteration = iteration + 1

      }

    } while (!converged)
    
    */

    spark.stop()
  }


}
